/******************************************************************************
**  dls converter
**  --------------------------------------------------------------------------
**  
**  convert dls file into wave table header and a pcm / adpcm file.
**  
**  Author: s.taffs
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License as published by the
** Free Software Foundation; either version 2 of the License,
** or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
** See the GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.
**
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "types.h"
#include "dls.h"
#include "grok.h"

/* macro defines */

/* #define    debug    1 */

#define SWAP_16(a)(       \
            ((a & 0xFF00) >> 8) |   \
            ((a & 0x00FF) << 8))

#define SWAP_32(a)(               \
            ((a & 0xFF000000) >> 24)    |   \
            ((a & 0x00FF0000) >> 8)     |   \
            ((a & 0x0000FF00) << 8)     |   \
            ((a & 0x000000FF) << 24))

#define    F_WSMP_NO_COMPRESSION 0x0002

#define    PACKET_SAMPLES    14
#define    PACKET_BYTES      8
#define    MIN(a,b) (((a)<(b))?(a):(b))

/* external variables */

/* None */

/* local variables */

static FILE *dls_file_ptr;
static FILE *wt_file_ptr;
static FILE *pcm_file_ptr;

static S32  current_program;
static S32  percussive;

static WT_INST_TYPE   melodic_inst[128];
static WT_INST_TYPE   percussive_inst[128];

static U16           art_index;
static WT_ART_TYPE   art[0xFFFF];
static WT_ART_TYPE   art_temp;

static U16            region_index;
static WT_REGION_TYPE region[0xFFFF];

static S32            wave_offset;
static S32            wave_index;
static WT_SAMPLE_TYPE wave_sample[0xFFFF]; 

static U16            adpcm_index;
static WT_ADPCM_TYPE  adpcm[0xFFFF];

static U32      compression;
static U32      loop_start;
static U32      loop_length;
static U32      encode_adpcm;
static U32      convert_to_16bit;

/* global variables */

/* None */

/* local prototypes */

static void dls_string(void);
static void dls_skip_chunk(void);
static void dls_riff(void);
static void dls_colh(void);
static void dls_vers(void);
static void dls_list(void);
static void dls_insh(void);
static void dls_rgnh(void);
static void dls_wsmp(void);
static void dls_wlnk(void);
static void dls_fmt(void);
static void dls_data(void);
static void dls_art1(void);
static void dls_art_default(void);
static void dls_connection(U16 source, U16 control, U16 destination, S32 scale);
static void dls_set_art(void);
static S32 dls_get_pan(S32 scale);
static S32 dls_get_eg1_sustain(S32 scale);
static U32 dls_tc2ms(S32 scale);
static S32 dls_get_lfo_freq(S32 scale);
static S32 dls_get_lfo_delay(S32 scale);
static S32 dls_get_eg1_release(S32 scale);
static S32 dls_get_eg2_sustain(S32 scale, S32 cents);
static S32 dls_get_eg2_release(S32 scale, S32 cents);
static S32 dls_get_bytes_for_adpcm(S32 samples);

/*
================================================================
=
= Function: dls_read_file
=
= Description:
=
=    convert the dls file into wave header and pcm file 
=
================================================================ 
*/
void dls_read_file(FILE *dls_file, FILE *wt_file, FILE *pcm_file, S32 mode)
{
	WT_FILE_HEADER_TYPE file_header;
	U32 chunk = 0;
	U32 i;
	U32 j;
	S32 file_position;
	
    /* initialize local file pointers */
    dls_file_ptr         = dls_file;
    wt_file_ptr          = wt_file;
    pcm_file_ptr         = pcm_file;
    
    /* init the index's*/
    region_index     = 0;
    art_index        = 0;
    wave_index       = 0;
    wave_offset      = 0;
    adpcm_index      = 0;
    
    /* set the compression type */
    compression     = mode;
    
    /* mark all instrument regions invalid */
    for (i = 0; i < 128; i++)
    {
        for (j = 0; j < 128; j++)
        {
            percussive_inst[i].key_region[j] = 0xFFFF;
        }
    }

    for (i = 0; i < 128; i++)
    {
        for (j = 0; j < 128; j++)
        {
            melodic_inst[i].key_region[j] = 0xFFFF;
        }
    }
	
	/* loop until we have converted all the dls file */
	while (1)
	{
		chunk = 0;

        /* read in the next chunk */
        if (fread(&chunk, 1, 4, dls_file_ptr) == 0)
        {
            printf("\n end of DLS file! \n");
            break;
        }
        
        switch(chunk)
		{
            case RIFF:  
			    dls_riff(); 
			break;
			
		    case COLH:  
			    dls_colh();
			break;
			
		    case VERS:  
			    dls_vers(); 
			break;
			
            case LIST:  
			    dls_list(); 
			break;
			
            case INSH:  
			    dls_insh(); 
			break;
			
            case RGNH:  
			    dls_rgnh(); 
			break;
			
            case WSMP:  
			    dls_wsmp(); 
			break;
			
            case WLNK:  
			    dls_wlnk(); 
			break;
			
            case ART1:  
			    dls_art1(); 
			break;
			
            case FMT:   
			    dls_fmt();  
			break;
			
            case DATA:  
			    dls_data(); 
			break;
        
            case DLS:
        	    printf("\n start of DLS file! \n");
		    break;

            /* information string chunks */
			case IARL:
            case IART:
            case ICMS:
            case ICMT:
            case ICOP:
            case ICRD:
            case IENG:
            case IGNR:
            case IKEY:
            case IMED:
            case INAM:
            case IPRD:
            case ISBJ:
            case ISFT:
            case ISRC:
            case ISRF:
            case ITCH:
                dls_string();       
            break;

            case PTBL:
		    default:
		    	/* skip unknown / unused chunks */
                dls_skip_chunk();
			break;
		}
    } 
   
    /* write wt file */
    fwrite(&file_header, sizeof(WT_FILE_HEADER_TYPE), 1, wt_file_ptr);
    
    file_position = sizeof(WT_FILE_HEADER_TYPE);
    file_header.offset_percussive_inst = SWAP_32(file_position);
    
    fwrite(percussive_inst, sizeof(WT_INST_TYPE), 128, wt_file_ptr);

    file_position += sizeof(WT_INST_TYPE) * 128;
    file_header.offset_melodic_inst    = SWAP_32(file_position);

    fwrite(melodic_inst, sizeof(WT_INST_TYPE), 128, wt_file_ptr);

    file_position += sizeof(WT_INST_TYPE) * 128;
    file_header.offset_regions        = SWAP_32(file_position);

    fwrite(region, sizeof(WT_REGION_TYPE), region_index, wt_file_ptr);
    
    file_position += sizeof(WT_REGION_TYPE) * region_index;
    file_header.offset_articulations  = SWAP_32(file_position);

    fwrite(art, sizeof(WT_ART_TYPE), art_index, wt_file_ptr);

    file_position += sizeof(WT_ART_TYPE) * art_index;
    file_header.offset_samples        = SWAP_32(file_position);

    fwrite(wave_sample, sizeof(WT_SAMPLE_TYPE), wave_index, wt_file_ptr);

    file_position += sizeof(WT_SAMPLE_TYPE) * wave_index;
    file_header.offset_adpcm_context   = SWAP_32(file_position);
    
    fwrite(adpcm, sizeof(WT_ADPCM_TYPE), adpcm_index, wt_file_ptr);

    /* write the real header again */
    fseek(wt_file_ptr, 0, SEEK_SET);
    fwrite(&file_header, sizeof(WT_FILE_HEADER_TYPE), 1, wt_file_ptr);
}

/*
================================================================
=
= Function: dls_string
=
= Description:
=
=    read and display dls info strings 
=
================================================================ 
*/
static void dls_string(void)
{
    U32 count;
    char ch;

    fread(&count, 1, 4, dls_file_ptr);
    
    /* fix alignment */
    if (count % sizeof(U16))
    {
        count += 1;
    }

    while (count)
    {
        fread(&ch, 1, 1, dls_file_ptr);
#ifdef debug
        printf("%c", ch);
#endif    
        count--;
    }
#ifdef debug
    printf("\n");
#endif  
}

/*
================================================================
=
= Function: dls_skip_chunk
=
= Description:
=
=    skip unknwon / unused chunks
=
================================================================ 
*/
static void dls_skip_chunk(void)
{
    U32 count;

    fread(&count, 1, 4, dls_file_ptr);
    
    /* fix alignment */
    if (count % sizeof(U16))
    {
        count += 1;
    }

    fseek(dls_file_ptr, count, SEEK_CUR);
#ifdef debug    
    printf("Skipping %d bytes\n", count);
#endif  
}

/*
================================================================
=
= Function: dls_riff
=
= Description:
=
=    read the total dls file size
=
================================================================ 
*/
static void dls_riff(void)
{
    U32 count;

    fread(&count, 1, 4, dls_file_ptr);
    
    printf(" dls file size %d bytes\n", count);
}

/*
================================================================
=
= Function: dls_colh
=
= Description:
=
=    read the collection header chunk
=
================================================================ 
*/
static void dls_colh(void)
{
    U32 count;
	U32 length;

    fread(&length, 1, 4, dls_file_ptr);
    fread(&count, 1, 4, dls_file_ptr);

    /* check to see if we have advanced the read position by specified length */
    length -= 4;

    if(length != 0)
    {
        fseek(dls_file_ptr, length, SEEK_CUR);
    }
    
    printf(" file contains %d instruments\n", count);
}

/*
================================================================
=
= Function: dls_vers
=
= Description:
=
=    read the dls file version
=
================================================================ 
*/
static void dls_vers(void)
{
    U32 ms;
	U32 ls; 
	U32 length;

    fread(&length, 1, 4, dls_file_ptr);
    fread(&ms, 1, 4, dls_file_ptr);
    fread(&ls, 1, 4, dls_file_ptr);

    /* check to see if we have advanced the read position by specified length */
    length -= 8;

    if(length != 0)
    {
        fseek(dls_file_ptr, length, SEEK_CUR);
    }

    printf(
        " file VERSION %d,%d,%d,%d\n",
        ms >> 16,
        ms & 0xFFFF,
        ls >> 16,
        ls & 0xFFFF
    );
}

/*
================================================================
=
= Function: dls_list
=
= Description:
=
=    read the list chunk type
=
================================================================ 
*/
static void dls_list(void)
{
    U32 list;

    fseek(dls_file_ptr, 4, SEEK_CUR);
    fread(&list, 1, 4, dls_file_ptr);
    
#ifdef debug    
    printf(
        "<%c%c%c%c>\n",
        (list & 0xFF),
        ((list >> 8) & 0xFF),
        ((list >> 16) & 0xFF),
        ((list >> 24) & 0xFF) 
        );
#endif 
}

/*
================================================================
=
= Function: dls_insh
=
= Description:
=
=    read the instrument header chunk
=
================================================================ 
*/
static void dls_insh(void)
{
    U32 regions;
    U32 bank;
    U32 program;
    U32 length;

    fread(&length, 1, 4, dls_file_ptr);
    fread(&regions, 1, 4, dls_file_ptr);
    fread(&bank, 1, 4, dls_file_ptr);
    fread(&program, 1, 4, dls_file_ptr);

    /* check to see if we have advanced the read position by specified length */
    length -= 12;

    if(length != 0)
    {
        fseek(dls_file_ptr, length, SEEK_CUR);
    }

#ifdef debug 
    printf("%d regions, bank: %d, program: %d\n", regions, bank, program);
#endif
    
    /* check bit 31 to see if instrument is drum or melodic */
    if (bank & 0x80000000)
    {
        percussive = 1;    /* instrument is a drum instrument */
    }
    else
    {
        percussive = 0;    /* melodic instrument */
    }

    /* this wt format does not support banks so if the instrument is not on */
    /* bank 0 do not make any instrument reference to it                    */
    if (bank & 0x0000FFFF)
    {
        current_program = 0xFFFFFFFF;
    }
    else
    {
        current_program = program;
    }
}

/*
================================================================
=
= Function: dls_rgnh
=
= Description:
=
=    read the instruments region key info
=
================================================================ 
*/
static void dls_rgnh(void)
{
    U16         lo_key;
    U16         hi_key;
    U16         lo_vel;
    U16         hi_vel;
    U16         options;
    U16         key_group;
    WT_REGION_TYPE    *this_region;
    U32         length;
    S32         i;

    fread(&length, 1, 4, dls_file_ptr);
    fread(&lo_key, 1, 2, dls_file_ptr);
    fread(&hi_key, 1, 2, dls_file_ptr);
    fread(&lo_vel, 1, 2, dls_file_ptr);
    fread(&hi_vel, 1, 2, dls_file_ptr);
    fread(&options, 1, 2, dls_file_ptr);
    fread(&key_group, 1, 2, dls_file_ptr);
    
#ifdef debug  
    printf("loKey: %d, hiKey: %d\n", lo_key, hi_key);
    printf("loVel: %d, hiVel: %d\n", lo_vel, hi_vel);
    printf("Options: 0x%.4x\n", options);
    printf("Key_group: %d\n", key_group);
#endif

    /* check to see if we have advanced the read position by specified length */
    length -= 12;

    if(length != 0)
    {
        fseek(dls_file_ptr, length, SEEK_CUR);
    }

    this_region = &region[region_index];

    this_region->key_group    = (U8)key_group;

    /* mark the current instrument's keys for this region */
    if (current_program != 0xFFFFFFFF)
    {
        if (percussive)
        {
            for (i = lo_key; i <= hi_key; i++)
            {
                if (percussive_inst[current_program].key_region[i] == 0xFFFF)
                {
                    percussive_inst[current_program].key_region[i] = SWAP_16(region_index);
                }
            }
        }
        else
        {
            for (i = lo_key; i <= hi_key; i++)
            {
                if (melodic_inst[current_program].key_region[i] == 0xFFFF)
                {
                    melodic_inst[current_program].key_region[i] = SWAP_16(region_index);
                }
            }
        }
    }
}

/*
================================================================
=
= Function: dls_wsmp
=
= Description:
=
=    wave sample chunk describes pitch / volume to play sample 
=    and loop info
=
================================================================ 
*/
static void dls_wsmp(void)
{
    U32         struct_size;
    U16         unity_note;
    S16         fine_tune;
    S32         attenuation;
    U32         options;
    U32         loops;

    U32         size;
    U32         type;
    U32         start;
    U32         length;

    WT_REGION_TYPE    *thisRegion;    

    fseek(dls_file_ptr, 4, SEEK_CUR);
    fread(&struct_size, 1, 4, dls_file_ptr);
    fread(&unity_note, 1, 2, dls_file_ptr);
    fread(&fine_tune, 1, 2, dls_file_ptr);
    fread(&attenuation, 1, 4, dls_file_ptr);
    fread(&options, 1, 4, dls_file_ptr);
    fread(&loops, 1, 4, dls_file_ptr);
   
#ifdef debug
    printf("Struct size: %d\n", struct_size);
    printf("unity_note: %d\n", unity_note);
    printf("fine_tune: 0x%.4x\n", fine_tune);
    printf("Attenuation: 0x%.8x\n", attenuation);
    printf("Options: 0x%.8x\n", options);
    printf("Loops: %d\n", loops);
#endif

    thisRegion = &region[region_index];

    thisRegion->unity_note   = (U8)unity_note;

    thisRegion->fine_tune   = SWAP_16(fine_tune);
    thisRegion->attn        = SWAP_32(attenuation);
    
    if (loops)
    {
        fread(&size, 1, 4, dls_file_ptr);
        fread(&type, 1, 4, dls_file_ptr);
        fread(&start, 1, 4, dls_file_ptr);
        fread(&length, 1, 4, dls_file_ptr);
#ifdef debug
        printf("\n");
        printf("size: %d\n", size);
        printf("type: 0x%.8x\n", type);
        printf("start: %d\n", start);
        printf("length: %d\n", length);
#endif
        thisRegion->loop_start  = SWAP_32(start);
        thisRegion->loop_length = SWAP_32(length);
        loop_start   = start;
        loop_length  = length;
    }
    else
    {
        thisRegion->loop_start   = 0;
        thisRegion->loop_length  = 0;
    }

    switch (compression)
    {
        case MODE_USE_FLAG: /* use F_WSMP_NO_COMPRESSION flag */
        	if (options & F_WSMP_NO_COMPRESSION)
        	{
            	encode_adpcm = 0;
        	}
        	else
        	{
            	encode_adpcm = 1;
        	}
        break;

        case MODE_COMPRESS_ALL:
            encode_adpcm = 1;
        break;
    
        case MODE_COMPRESS_NONE:
            encode_adpcm = 0;
        break;
    }
}

/*
================================================================
=
= Function: dls_wlnk
=
= Description:
=
=    specifies where the wave data can be found
=
================================================================ 
*/
static void dls_wlnk(void)
{
    U16         options;
    U16         phase_group;
    U32         channel;
    U32         table_index;
    WT_REGION_TYPE    *this_region;

    fseek(dls_file_ptr, 4, SEEK_CUR);
    fread(&options, 1, 2, dls_file_ptr);
    fread(&phase_group, 1, 2, dls_file_ptr);
    fread(&channel, 1, 4, dls_file_ptr);
    fread(&table_index, 1, 4, dls_file_ptr);

#ifdef debug
    printf("Options: 0x%.4x\n", options);
    printf("PhaseGroup: 0x%.4x\n", phase_group);
    printf("Channel: %d\n", channel);
    printf("Table_index: %d\n", table_index);
#endif

    this_region = &region[region_index];

    this_region->sample_index         = SWAP_32(table_index);
    this_region->articulation_index   = SWAP_32(art_index);  
    
    region_index++;    
}

/*
================================================================
=
= Function: dls_art1
=
= Description:
=
=    sLevel 1 Articulator Chunk
=
================================================================ 
*/
static void dls_art1(void)
{
    U32 struct_size;
    U32 connection_blocks;

    U16 source;
    U16 control;
    U16 destination;
    U16 transform;
    S32 scale;
    
    fseek(dls_file_ptr, 4, SEEK_CUR);
    fread(&struct_size, 1, 4, dls_file_ptr);
    fread(&connection_blocks, 1, 4, dls_file_ptr);
#ifdef debug
    printf("StructSize: %d\n", struct_size);
    printf("ConnectionBlocks: %d", connection_blocks);
#endif
    
    dls_art_default();

    while (connection_blocks)
    {
        fread(&source, 1, 2, dls_file_ptr);
        fread(&control, 1, 2, dls_file_ptr);
        fread(&destination, 1, 2, dls_file_ptr);
        fread(&transform, 1, 2, dls_file_ptr);
        fread(&scale, 1, 4, dls_file_ptr);
#ifdef debug
        printf("\n");
        printf("Source: 0x%.4x\n", source);
        printf("Control: 0x%.4x\n", control);
        printf("Destination: 0x%.4x\n", destination);
        printf("Transform: 0x%.4x\n", transform);
        printf("Scale: 0x%.8x\n", scale);
#endif
        dls_connection(source, control, destination, scale);
        
        connection_blocks--;
    }

    dls_set_art();    
    art_index++;
}

/*
================================================================
=
= Function: dls_fmt
=
= Description:
=
=    specifies wave data format
=
================================================================ 
*/
static void dls_fmt(void)
{
    U16 format_tag;
    U16 channels;
    U32 samples_per_sec;
    U32 average_bytes_per_sec;
    U16 block_align;
    U16 bits_per_sample;
	U16 samples_per_sec16;
    U16 format;

    fseek(dls_file_ptr, 4, SEEK_CUR);
    fread(&format_tag, 1, 2, dls_file_ptr);
    fread(&channels, 1, 2, dls_file_ptr);
    fread(&samples_per_sec, 1, 4, dls_file_ptr);
    fread(&average_bytes_per_sec, 1, 4, dls_file_ptr);
    fread(&block_align, 1, 2, dls_file_ptr);
    fread(&bits_per_sample, 1, 2, dls_file_ptr);
    fseek(dls_file_ptr, 2, SEEK_CUR);

#ifdef debug
    printf("FormatTag: 0x%.4x\n", format_tag);
    printf("Channels: %d\n", channels);
    printf("SamplesPerSec: %d\n", samples_per_sec);
    printf("AverageBytesPerSec: %d\n", average_bytes_per_sec);
    printf("BlockAlign: %d\n", block_align);
    printf("BitsPerSample: %d\n", bits_per_sample);
#endif

    /* give warning if format tag is not PCM */
    if (format_tag != 0x0001)
    {
        printf("Warning: sample %d is not in PCM format!%c\n", wave_index,7);
    }

    switch (bits_per_sample)
    {
        case 16:
            if (encode_adpcm == 1)
            {
                format = WT_FORMAT_ADPCM;
                convert_to_16bit = 0;
            }
            else
            {
                format = WT_FORMAT_PCM16;
            }
        break;

        case 8:
            if(encode_adpcm == 1)
            {
                format = WT_FORMAT_ADPCM;
                convert_to_16bit = 1;
            }
            else
            {
                format = WT_FORMAT_PCM8;
            }
        break;

        default:
            printf("Warning: sample %d is not 16 or 8 bits per sample!%c\n", wave_index, 7 );
    }

	samples_per_sec16 = (U16)samples_per_sec;
    wave_sample[wave_index].format        = SWAP_16(format);
    wave_sample[wave_index].sample_rate   = SWAP_16(samples_per_sec16);
}

/*
================================================================
=
= Function: dls_data
=
= Description:
=
=    specifies wave data and its size
=
================================================================ 
*/
static void dls_data(void)
{
    U32 count;
    U8 sample_8bit;
    U16 sample_16bit;
    U8 bytes;
    U8 ch;
    
    S32 i;
	S32 p;
	S32 s;
    WT_ADPCM_TYPE adpcm_header;
    S16 *pcm_buffer = NULL;
    U32 samples = 0;
    S16 coefs[16];
    S32 packet_count = 0;
    S32 loop_packet = 0;
    S16 conv_samples[16];
	U8 block[8];
	S32 num_samples = 0;
	U8 sample = 0;
	S16 *ptr_buf = NULL;
	U32 temp_count = 0;
	U8 sample_pos_in_frame;

    fread(&count, 1, 4, dls_file_ptr);

    switch (SWAP_16(wave_sample[wave_index].format))
    {
        case WT_FORMAT_ADPCM:

            /* ADPCM has to start on a frame boundry (8 bytes) */
            bytes = wave_offset % 8;

            if(bytes != 0)
            {
                bytes = 8 - bytes;

                while (bytes)
                {
                    ch = 0;
                    fwrite(&ch, 1, 1, pcm_file_ptr);
                    wave_offset++;
                    bytes--;
                }
            }

            samples = count / 2;
            pcm_buffer = malloc(count);
            packet_count = samples / PACKET_SAMPLES + (samples % PACKET_SAMPLES != 0);

            if((loop_start + loop_length) != 0)
            {
                loop_packet = (loop_start - 1) / PACKET_SAMPLES + ((loop_start -1) % PACKET_SAMPLES != 0);
            }

            wave_sample[wave_index].length = SWAP_32(samples);
            wave_sample[wave_index].offset = SWAP_32(wave_offset * 2);    

            if(pcm_buffer != NULL)
            {
            	memset(conv_samples, 0, (PACKET_SAMPLES + 2) * sizeof(S16));
                memset(&adpcm_header, 0, sizeof(WT_ADPCM_TYPE));
                    
                /* see if the sample needs to be converted to 16 bit first */
				if (1 == convert_to_16bit)
				{
					temp_count = count;
					ptr_buf = pcm_buffer;

                    while(temp_count != 0)
                    {
                        fread(&sample, 1, 1, dls_file_ptr);
                        *ptr_buf++ = (S16)((sample + 0x80) << 8);
                
                        temp_count--;
                    }
				}
				else
				{
					fread(pcm_buffer, count, 1, dls_file_ptr);
				}
            	
                /* find the best fit coef for this sample */
				DSPCorrelateCoefs(pcm_buffer, samples, coefs);
					
				for(i = 0; i < 16 ; i++)
				{
					adpcm_header.coef[i] = SWAP_16(coefs[i]);
				}
				    
				/* Execute encoding-predictor for each block */
				for(p = 0 ; p < packet_count ; p++)
    			{
        			memset(conv_samples + 2, 0, PACKET_SAMPLES * sizeof(S16));
        			num_samples = MIN(samples - p * PACKET_SAMPLES, PACKET_SAMPLES);

        			for(s = 0 ; s < num_samples; s++)
        			{
            			conv_samples[s + 2] = pcm_buffer[p * PACKET_SAMPLES + s];
            		}

        			DSPEncodeFrame(conv_samples, PACKET_SAMPLES, block, coefs);
        			
        			/* save the first pred / scale block in the header */
        			if(p == 0)
        			{
        				adpcm_header.pred_scale = SWAP_16((U16)block[0]);
					}
        			
        			if((loop_start + loop_length) != 0)
        			{
        			    if((loop_packet - 1) == p)
        			    {
        			    	/* store the loop scale and coefficient values in header */
        				    adpcm_header.loop_pred_scale = SWAP_16((U16)block[0]);
					        
					        sample_pos_in_frame = ((loop_start - 1) % PACKET_SAMPLES);
					    
					        if(sample_pos_in_frame == 0)
					        {
					        	adpcm_header.loop_yn2 = SWAP_16((U16)conv_samples[0]);
					    	    adpcm_header.loop_yn1 = SWAP_16((U16)conv_samples[1]);
						    }
						    else if(sample_pos_in_frame == 1)
						    {
						    	adpcm_header.loop_yn2 = SWAP_16((U16)conv_samples[1]);
								adpcm_header.loop_yn1 = SWAP_16((U16)conv_samples[2]);
						    }
						    else
						    {
						    	adpcm_header.loop_yn2 = SWAP_16((U16)conv_samples[2 + sample_pos_in_frame -2]);
						    	adpcm_header.loop_yn1 = SWAP_16((U16)conv_samples[2 + sample_pos_in_frame -1]);
							}
					    }
				    }
        			
        			fwrite(block, 1, dls_get_bytes_for_adpcm(num_samples), pcm_file_ptr);
        			wave_offset += dls_get_bytes_for_adpcm(num_samples);

                    /* y[n - 1] and y[n - 2] */
        			conv_samples[0] = conv_samples[14];
					conv_samples[1] = conv_samples[15];
                }
                
                /* save the header before processing the next adpcm waveform */
				memcpy(&adpcm[adpcm_index], &adpcm_header, sizeof(WT_ADPCM_TYPE));
            }
            else
            {
                printf("Warning: could not allocate buffer for ADPCM encode!\n%c", 7);
            }
                
            if(pcm_buffer != NULL)
			{
				free(pcm_buffer);
			}

            wave_sample[wave_index].adpcm_index = SWAP_16(adpcm_index);
            adpcm_index++;

        break;

        case WT_FORMAT_PCM16:

            wave_sample[wave_index].length = SWAP_32(count / 2);
            wave_sample[wave_index].offset = SWAP_32(wave_offset / 2);    
            wave_offset += count;

			/* write the wave data output to the pcm file */
            while(count)
            {        
                fread(&sample_16bit, 1, 2, dls_file_ptr);
                sample_16bit = SWAP_16(sample_16bit);
                fwrite(&sample_16bit, 1, 2, pcm_file_ptr);

                count -= 2;
            }

        break;

        case WT_FORMAT_PCM8:

            wave_sample[wave_index].length = SWAP_32(count);
            wave_sample[wave_index].offset = SWAP_32(wave_offset);    
            wave_offset += count;
    
            /* write the wave data output to the pcm file */
            while(count)
            {       
                fread(&sample_8bit, 1, 1, dls_file_ptr);
                sample_8bit += 0x80;
                fwrite(&sample_8bit, 1, 1, pcm_file_ptr);

                count--;
            }

        break;
    }

    wave_index++;
    printf(".");
}

/*
================================================================
=
= Function: dls_get_bytes_for_adpcm
=
= Description:
=
=    returns the amount of compressed space required for the
=    amount of uncompressed samples passed
=
================================================================ 
*/
static S32 dls_get_bytes_for_adpcm(S32 samples)
{
    S32 extra_bytes = 0;
    S32 packets = samples / PACKET_SAMPLES;
    S32 extra_samples = samples % PACKET_SAMPLES;

    if (extra_samples != 0)
    {
        extra_bytes = (extra_samples / 2) + (extra_samples % 2) + 1;
    }

    return PACKET_BYTES * packets + extra_bytes;
}

/*
================================================================
=
= Function: dls_art_default
=
= Description:
=
=    default articulator setup
=
================================================================ 
*/
static void dls_art_default(void)
{
    WT_ART_TYPE *art1 = &art_temp;

    art1->lfo_freq          = 0xFCACAE9C;       /* 5Hz        */
    art1->lfo_delay         = 0xE0DB6022;       /* 0.01 sec   */
    art1->lfo_atten         = 0x00000000;       /* 0dB        */
    art1->lfo_pitch         = 0x00000000;       /* 0 cents    */
    art1->lfo_mod2atten     = 0x00000000;       /* 0dB        */
    art1->lfo_mod2pitch     = 0x00000000;       /* 0 cents    */

    art1->eg1_attack        = 0x80000000;       /* 0 seconds  */
    art1->eg1_decay         = 0x80000000;       /* 0 seconds  */ 
    art1->eg1_sustain       = 0x03E80000;       /* 100%       */
    art1->eg1_release       = 0x80000000;       /* 0 seconds  */
    art1->eg1_vel2attack    = 0x80000000;       /* 0 seconds  */
    art1->eg1_key2decay     = 0x80000000;       /* 0 seconds  */

    art1->eg2_attack        = 0x80000000;       /* 0 seconds  */
    art1->eg2_decay         = 0x80000000;       /* 0 seconds  */
    art1->eg2_sustain       = 0x03E80000;       /* 100%       */
    art1->eg2_release       = 0x80000000;       /* 0 seconds  */
    art1->eg2_vel2attack    = 0x80000000;       /* 0 seconds  */
    art1->eg2_key2decay     = 0x80000000;       /* 0 seconds  */
    art1->eg2_pitch         = 0x00000000;       /* 0 cents    */

    art1->pan               = 0x00000000;       /* center     */ 
}

/*
================================================================
=
= Function: dls_connection
=
= Description:
=
=    work out scale art1 connection
=
================================================================ 
*/
static void dls_connection (U16 source, U16 control, U16 destination, S32 scale)
{					
    WT_ART_TYPE *art1 = &art_temp;

    switch (source)
	{
	    case CONN_SRC_NONE:		
	        switch (destination)
		    {
		        case CONN_DST_LFO_FREQUENCY:    art1->lfo_freq      = scale;    break;
		        case CONN_DST_LFO_STARTDELAY:   art1->lfo_delay     = scale;    break;
		        case CONN_DST_EG1_ATTACKTIME:   art1->eg1_attack    = scale;    break;
		        case CONN_DST_EG1_DECAYTIME:    art1->eg1_decay     = scale;    break;
		        case CONN_DST_EG1_SUSTAINLEVEL: art1->eg1_sustain   = scale;    break;
		        case CONN_DST_EG1_RELEASETIME:	art1->eg1_release   = scale;    break;
		        case CONN_DST_EG2_ATTACKTIME:	art1->eg2_attack    = scale;    break;
	 	        case CONN_DST_EG2_DECAYTIME:	art1->eg2_decay     = scale;    break;
		        case CONN_DST_EG2_SUSTAINLEVEL:	art1->eg2_sustain   = scale;    break;
		        case CONN_DST_EG2_RELEASETIME:	art1->eg2_release   = scale;    break;
		        case CONN_DST_EG1_RESERVED:                                     break;
		        case CONN_DST_EG2_RESERVED:                                     break;	
		        case CONN_DST_PAN:              art1->pan           = scale;    break;
		    } 				
        break;
						
	    case CONN_SRC_LFO:
		    switch (destination)
		    {
		        case CONN_DST_ATTENUATION:							
			        switch (control)
			        {
			            case CONN_SRC_NONE:         art1->lfo_atten     = scale;    break;
			            case CONN_SRC_CC1:			art1->lfo_mod2atten = scale;    break;
			        }						
			    break;
				
		        case CONN_DST_PITCH:	
			        switch (control)
			        {
			            case CONN_SRC_NONE:         art1->lfo_pitch     = scale;    break;
			            case CONN_SRC_CC1:			art1->lfo_mod2pitch = scale;    break;
			        }		
			    break;
		    }
		break;
						
	    case CONN_SRC_KEYONVELOCITY:			
		    switch (destination)
		    {
		        case CONN_DST_EG1_ATTACKTIME:   art1->eg1_vel2attack = scale;   break;
		        case CONN_DST_EG2_ATTACKTIME:	art1->eg2_vel2attack = scale;   break;   
		    }			
		break;
						
	    case CONN_SRC_KEYNUMBER:			
		    switch (destination)
		    {
		        case CONN_DST_EG1_DECAYTIME:	art1->eg1_key2decay = scale;    break;
		        case CONN_DST_EG2_DECAYTIME:	art1->eg2_key2decay = scale;    break;
		    }
		break;
						
	    case CONN_SRC_EG2:			
		    switch (destination)
		    {
		        case CONN_DST_PITCH:            art1->eg2_pitch     = scale;    break;
		    }
		break;		 		
	}
}

/*
================================================================
=
= Function: dls_set_art
=
= Description:
=
=    write out final art1 values to art array
=
================================================================ 
*/
static void dls_set_art(void)
{
    WT_ART_TYPE *art1 = &art[art_index];

    /* write out LFO */
    art1->lfo_freq          = SWAP_32(dls_get_lfo_freq(art_temp.lfo_freq));  
    art1->lfo_delay         = SWAP_32(dls_get_lfo_delay(art_temp.lfo_delay));
    art1->lfo_atten         = SWAP_32(art_temp.lfo_atten);
    art1->lfo_pitch         = SWAP_32(art_temp.lfo_pitch);
    art1->lfo_mod2atten     = SWAP_32(art_temp.lfo_mod2atten);
    art1->lfo_mod2pitch     = SWAP_32(art_temp.lfo_mod2pitch);

    /* write out (eg1) volume envelope */
    art1->eg1_attack        = SWAP_32(art_temp.eg1_attack);
    art1->eg1_decay         = SWAP_32(art_temp.eg1_decay);
    art1->eg1_sustain       = SWAP_32(dls_get_eg1_sustain(art_temp.eg1_sustain));    
    art1->eg1_release       = SWAP_32(dls_get_eg1_release(art_temp.eg1_release));
    art1->eg1_vel2attack    = SWAP_32(art_temp.eg1_vel2attack);
    art1->eg1_key2decay     = SWAP_32(art_temp.eg1_key2decay);

    /* write out (eg2) pitch envelope */
    art1->eg2_pitch         = SWAP_32(art_temp.eg2_pitch);         
    art1->eg2_attack        = SWAP_32(art_temp.eg2_attack);
    art1->eg2_decay         = SWAP_32(art_temp.eg2_decay);
    art1->eg2_sustain       = SWAP_32(dls_get_eg2_sustain(art_temp.eg2_sustain, art_temp.eg2_pitch));    
    art1->eg2_release       = SWAP_32(dls_get_eg2_release(art_temp.eg2_release, art_temp.eg2_pitch));
    art1->eg2_vel2attack    = SWAP_32(art_temp.eg2_vel2attack);
    art1->eg2_key2decay     = SWAP_32(art_temp.eg2_key2decay);

    /* write out pan */
    art1->pan              = SWAP_32(dls_get_pan(art_temp.pan));
}

/*
================================================================
=
= Function: dls_tc2ms
=
= Description:
=
=    time cents to milliseconds
=
================================================================ 
*/
static U32 dls_tc2ms(S32 scale)    
{
	if (scale == 0x80000000)
	{
		return 0;
    }

    return (U32)(pow(2, (double)scale / (1200 * 65536)) * 1000);            
}

/*
================================================================
=
= Function: dls_get_lfo_freq
=
= Description:
=
= convert pitch cents to frequency then works out the step
= size required for the lfo to run at the requested freq
=
================================================================ 
*/
static S32 dls_get_lfo_freq(S32 scale)
{
    float f;
    
    /* pitch cent to frequency using the follow formula */
    /* f = 2^((pc/65536 - 6900)/1200) * 440             */
    /* where pc = pitch Cents (scale in this case)      */
    f = (float)pow(2, ((double)scale / 65536 - 6900) / 1200) * 440;

    /* sine wave on the gc is running at 1 / 1ms = 1000hz      */
	/* the sine wave is broken into 64 steps so the sine wave  */ 
	/* completes 5ms * 64 steps every audio frame              */
	
	/* therefore we need a step size of (1000hz / f) / (5 * 64) */
	/* for the lfo to run at the requested rate                 */
    return (S32)(((1000.0f / f) / (5 * 64)) * 0x00010000); 
}

/*
================================================================
=
= Function: dls_get_lfo_delay
=
= Description:
=
=    
=
================================================================ 
*/
static S32 dls_get_lfo_delay(S32 scale)
{
    return dls_tc2ms(scale) / 65536;
}

/*
================================================================
=
= Function: dls_get_eg1_sustain
=
= Description:
=
=    
=
================================================================ 
*/
static S32 dls_get_eg1_sustain(S32 scale)
{
	/* 1% = 65535 * 10, so 100% is (65535 * 10) * 100 = 0x03E80000 */
    float percent = (float)scale / 0x03E80000;

    /* (960 << 16) = 0x03C00000, (-960 << 16) = 0xFC400000     */ 
	/* (-960 is the start) and zero is the finish -96 to 0 db  */
    return (S32)(0x03C00000 * percent) + 0xFC400000;
}

/*
================================================================
=
= Function: dls_get_eg1_release
=
= Description:
=
=    
=
================================================================ 
*/
static S32 dls_get_eg1_release(S32 scale)
{
    S32 frames;

	frames = dls_tc2ms(scale) / 5;	/* 5 ms per audio frame */
				
    if (frames)
    {
    	/* -960 = 96db so this calculates the how many frames it    */
    	/* will take the release value to get to -96b (zero volume) */
        return (S32)((-960 * 0x00010000) / frames);
    }
	else
	{
		/* release rate is 100% ie switch off straight away */
        return (S32)(-960 * 0x00010000);
    }
}

/*
================================================================
=
= Function: dls_get_eg2_sustain
=
= Description:
=
=    
=
================================================================ 
*/
static S32 dls_get_eg2_sustain(S32 scale, S32 cents)
{
    if (scale == 0x00000000)
    {
        return 0;
    }
    else
    {
        return (S32)(cents * ((double)scale / (1000 * 65536)));
    }
}

/*
================================================================
=
= Function: dls_get_eg2_release
=
= Description:
=
=    
=
================================================================ 
*/
static S32 dls_get_eg2_release(S32 scale, S32 cents)
{
    U32 frames;

	frames = dls_tc2ms(scale) / 5;	/* 5 ms per audio frame */
				
    if (frames)
    {
        return (S32)((cents * -1) / frames);
    }
	else
	{
        return cents * -1;
    }
}

/*
================================================================
=
= Function: dls_get_pan
=
= Description:
=
=    Pan Control (see figure 19 of DLS spec) 
=
================================================================ 
*/
static S32 dls_get_pan(S32 scale)
{
    if (scale == 0)
    {
        return 64;
    }

    return (S32)(127 * ((float)(scale + (500 * 0x00010000)) / (1000 * 0x00010000))) ;
}
